////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2021 Realm Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <exception>

#include "realm_dart.h"

#include <realm/object-store/c_api/types.hpp>
#include <realm/object-store/c_api/util.hpp>

#include <iostream>

namespace realm::c_api {

RLM_API void realm_dart_initializeDartApiDL(void* data) {
    Dart_InitializeApiDL(data);
}

static void handle_finalizer(void* isolate_callback_data, void* realmPtr) {
    realm_release(realmPtr);
}

RLM_API Dart_FinalizableHandle realm_dart_attach_finalizer(Dart_Handle handle, void* realmPtr, int size) {
    return Dart_NewFinalizableHandle_DL(handle, realmPtr, size, handle_finalizer);
}

RLM_API void realm_dart_delete_finalizable(Dart_FinalizableHandle finalizable_handle, Dart_Handle handle) {
    Dart_DeleteFinalizableHandle_DL(finalizable_handle, handle);
}

class WeakHandle {
public:
    WeakHandle(Dart_Handle handle) : m_weakHandle(Dart_NewWeakPersistentHandle_DL(handle, this, 1, finalize_handle)) {
    }

    Dart_Handle value() {
        return Dart_HandleFromWeakPersistent_DL(m_weakHandle);
    }

private:
    ~WeakHandle() {
        if (m_weakHandle) {
            Dart_DeleteWeakPersistentHandle_DL(m_weakHandle);
            m_weakHandle = nullptr;
        }
    }

    static void finalize_handle(void* isolate_callback_data, void* peer) {
        delete reinterpret_cast<WeakHandle*>(peer);
    }

    Dart_WeakPersistentHandle m_weakHandle;
};

RLM_API void* realm_dart_object_to_weak_handle(Dart_Handle handle) {
    return new WeakHandle(handle);
}

RLM_API Dart_Handle realm_dart_weak_handle_to_object(void* handle) {
    return reinterpret_cast<WeakHandle*>(handle)->value();
}

RLM_API void* realm_dart_object_to_persistent_handle(Dart_Handle handle) {
    return reinterpret_cast<void*>(Dart_NewPersistentHandle_DL(handle));
}

RLM_API Dart_Handle realm_dart_persistent_handle_to_object(void* handle) {
    Dart_PersistentHandle persistentHandle = reinterpret_cast<Dart_PersistentHandle>(handle);
    return Dart_HandleFromPersistent_DL(persistentHandle);
}

RLM_API void realm_dart_delete_persistent_handle(void* handle) {
    Dart_PersistentHandle persistentHandle = reinterpret_cast<Dart_PersistentHandle>(handle);
    Dart_DeletePersistentHandle_DL(persistentHandle);
}

namespace {

using FreeT = std::function<void()>;
using CallbackT = std::function<void(const realm_http_request_t request, void* request_context)>; // Differs per callback
using UserdataT = std::tuple<realm_scheduler_t*, CallbackT, FreeT>;


void _callback(realm_userdata_t userdata, const realm_http_request_t request, void* request_context) {
    // the pointers in error are to stack values, we need to make copies and move them into the scheduler invocation
    struct request_copy_buf {
        std::string url;
        std::string body;
        std::map<std::string, std::string> headers;
        std::vector<realm_http_header_t> headers_vector;
    } buf;

    buf.url = request.url;
    buf.body = std::string(request.body, request.body_size);
    for (size_t i = 0; i < request.num_headers; i++) {
        auto [it, _] = buf.headers.emplace(request.headers[i].name, request.headers[i].value);
        buf.headers_vector.push_back({ it->first.c_str(), it->second.c_str() });
    }

    realm_http_request_t request_copy = request; // copy struct
    request_copy.url = buf.url.c_str();
    request_copy.body = buf.body.data();
    request_copy.headers = buf.headers_vector.data();

    // trampoline to scheduler
    auto u = reinterpret_cast<UserdataT*>(userdata);
    auto scheduler = std::get<0>(*u);
    auto callback = std::get<1>(*u);
    (*scheduler)->invoke([buf = std::move(buf), request = std::move(request_copy), callback = std::move(callback), request_context]{
        callback(request, request_context);
    });
}

void _userdata_free(void* userdata) {
    auto u = reinterpret_cast<UserdataT*>(userdata);
    auto scheduler = std::get<0>(*u);
    auto userdata_free = std::get<2>(*u);
    (*scheduler)->invoke([userdata_free = std::move(userdata_free), u]{
        userdata_free();
        delete u;
    });
}

RLM_API realm_http_transport_t* realm_dart_http_transport_new(
    realm_http_request_func_t callback,
    realm_userdata_t userdata,
    realm_free_userdata_func_t userdata_free,
    realm_scheduler_t* scheduler)
{
    auto u = new UserdataT(scheduler,
                           std::bind(callback, userdata, std::placeholders::_1, std::placeholders::_2),
                           std::bind(userdata_free, userdata));
    return realm_http_transport_new(_callback, u, nullptr); //_userdata_free); // <-- crash on exit
}
} // anonymous namespace
} // namespace realm::c_api 
